/*************************************************************************\
* Copyright (c) 2006 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* This software is distributed subject to a Software License Agreement
* found in file LICENSE that is included with this distribution.
\*************************************************************************/

/*
 * Common Flash Interface flash memory programmer
 */

#include <bsp.h>
#include <string.h>
#include <epicsThread.h>
#include <epicsStdioRedirect.h>
#include <epicsStdlib.h>
#include <epicsString.h>
#include <epicsInterrupt.h>
#include <cantProceed.h>
#include <alarm.h>
#include <dbLock.h>
#include <devLib.h>
#include <devSup.h>
#include <iocsh.h>
#include <epicsExport.h>
#include "devFlashBurner.h"

struct eraseBlockInfo {
    unsigned int size;
    unsigned int count;
    unsigned int base;
};

struct flashInfo {
    volatile void        *base;
    int                   indirect;
    int                   byteMode;
    unsigned int          capacity;
    struct eraseBlockInfo eraseBlockInfo[4];
    double                sectorEraseTime;
    int                   sectorErasePassLimit;
    unsigned int          eraseHighWater;
};

static void
cfiReport(void *dpvt, FILE *fp, int level)
{
    struct flashInfo *f = dpvt;

    fprintf(fp, "   %d-bit, %sdirect, %d bytes\n", f->byteMode ? 8 : 16,
                                                   f->indirect ? "in" : "",
                                                   f->capacity);
}

/*
 * Compact Flash I/O
 */
struct indirectMaster {
    epicsUInt16 highAddress;
    epicsUInt16 lowAddress;
    epicsUInt16 data;
};

/*
 * GCC-3.4.3 improperly handles expression containing volatile when the
 * functions containing those expressions are inlined.
 */
static int readFlash(struct flashInfo *f, int addr) __attribute__((noinline));
static void writeFlash(struct flashInfo *f, int addr, int data) __attribute__((noinline));
static int
readFlash(struct flashInfo *f, int addr)
{
    if (f->indirect) {
        volatile struct indirectMaster *imp = f->base;
        imp->highAddress = addr >> 16;
        imp->lowAddress = addr;
        return imp->data;
    }
    else if (f->byteMode) {
        return *((volatile epicsUInt8 *)f->base + addr);
    }
    else {
        return *(volatile epicsUInt16 *)((char *)f->base + addr);
    }
}

static void
writeFlash(struct flashInfo *f, int addr, int data)
{
    if (f->indirect) {
        volatile struct indirectMaster *imp = f->base;
        imp->highAddress = addr >> 16;
        imp->lowAddress = addr;
        imp->data = data;
    }
    else if (f->byteMode) {
        *((volatile epicsUInt8 *)f->base + addr) = data;
    }
    else {
        *(volatile epicsUInt16 *)((char *)f->base + addr) = data;
    }
}

static asynStatus
getFlashInfo(struct flashInfo *f, int quickCheck)
{
    int i;
    epicsUInt8 q[0x51];
    epicsUInt8 manufacturerCode, deviceCode;
    unsigned int eraseBlockBase;
    int key;

    key = epicsInterruptLock();
    writeFlash(f, f->byteMode ? 0xAA : 0x55<<1, 0x98);
    for (i = 0 ; i < sizeof q / sizeof q[0] ; i++)
        q[i] = readFlash(f, i << 1);
    writeFlash(f, 0x00, 0xF0);
    epicsInterruptUnlock(key);
    if ((q[0x10] != 'Q') || (q[0x11] != 'R') || (q[0x12] != 'Y')) {
        printf("Flash Burner: Bad query response (%x %x %x).\n",
                                                    q[0x10], q[0x11], q[0x12]);
        return asynError;
    }
    if (quickCheck)
        return asynSuccess;
    key = epicsInterruptLock();
    writeFlash(f, f->byteMode ? 0xAAA : 0x555<<1, 0xAA);
    writeFlash(f, f->byteMode ? 0x555 : 0xAAA<<1, 0x55);
    writeFlash(f, f->byteMode ? 0xAAA : 0x555<<1, 0x90);
    manufacturerCode = readFlash(f, 0);
    deviceCode = readFlash(f, 2);
    writeFlash(f, 0x00, 0xF0);
    epicsInterruptUnlock(key);
    f->capacity = 1 << q[0x27];
    printf("Manufacturer Code:%#x  Device Code:%#x  Capacity:%#x\n",
                                    manufacturerCode, deviceCode, f->capacity);
    for (i = 0 ; i < 4 ; i++) {
        if (i < q[0x2C]) {
            f->eraseBlockInfo[i].size = (q[0x30 + 4*i]*256 + q[0x2F + 4*i]) * 256;
            f->eraseBlockInfo[i].count = (q[0x2E + 4*i]*256 + q[0x2D + 4*i]) + 1;
        }
        else {
            f->eraseBlockInfo[i].size = 0;
            f->eraseBlockInfo[i].count = 0;
        }
    }
    switch(manufacturerCode) {
    case 0x1F:          /* ATMEL */
        if ((deviceCode == 0xC8) /* AT49BV322A */
         && (f->eraseBlockInfo[0].size > f->eraseBlockInfo[1].size)) {
            struct eraseBlockInfo t;
            printf("ATMEL AT49BV322A -- Boot sectors at bottom -- reverse erase blocks.\n");
            t = f->eraseBlockInfo[0];
            f->eraseBlockInfo[0] = f->eraseBlockInfo[1];
            f->eraseBlockInfo[1] = t;
        }
        break;
    }
    eraseBlockBase = 0;
    for (i = 0 ; i < 4 ; i++) {
        if (f->eraseBlockInfo[i].size != 0) {
            f->eraseBlockInfo[i].base = eraseBlockBase;
            eraseBlockBase += f->eraseBlockInfo[i].count * f->eraseBlockInfo[i].size;
        printf("erase block %d size %#x count %d base %#x\n",
                                                    i,
                                                    f->eraseBlockInfo[i].size,
                                                    f->eraseBlockInfo[i].count,
                                                    f->eraseBlockInfo[i].base);
        }
    }
    if (eraseBlockBase != f->capacity)
        printf("Warning -- Flash memory capacity %d -- erase information for %d\n",
                                                   f->capacity, eraseBlockBase);
    f->sectorEraseTime = (1 << q[0x21]) / 1000.0;
    f->sectorErasePassLimit = 1 << q[0x25];
    printf("typical sector erase time %.3g seconds (max %.3g seconds)\n",
                                f->sectorEraseTime,
                                f->sectorEraseTime * f->sectorErasePassLimit);
    return asynSuccess;
}

static asynStatus
tryToErase(struct flashInfo *f, asynUser *pasynUser, unsigned int address)
{
    int d;
    int pass;
    int key;

    key = epicsInterruptLock();
    writeFlash(f, f->byteMode ? 0xAAA : 0x555<<1, 0xAA);
    writeFlash(f, f->byteMode ? 0x555 : 0xAAA<<1, 0x55);
    writeFlash(f, f->byteMode ? 0xAAA : 0x555<<1, 0x80);
    writeFlash(f, f->byteMode ? 0xAAA : 0x555<<1, 0xAA);
    writeFlash(f, f->byteMode ? 0x555 : 0xAAA<<1, 0x55);
    writeFlash(f, address, 0x30);
    epicsInterruptUnlock(key);
    for (pass = 0 ; ; pass++) {
        d = readFlash(f, address);
        if ((d & 0x80) == 0x80)
            return asynSuccess;
        if (pass > f->sectorErasePassLimit) {
            writeFlash(f, 0x0, 0xF0);
            epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                          "Timed out waiting for %#x erase (readback %#x).",
                                                                    address, d);
            return asynError;
        }
        if ((d & 0x20) != 0) {
            d = readFlash(f, address);
            if ((d & 0x80) == 0x80) 
                return asynSuccess;
            writeFlash(f, 0x0, 0xF0);
            epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                             "Erase %#x timed out (pass %d, readback %#x)",
                                                             address, pass, d);
            return asynError;
        }
        epicsThreadSleep(f->sectorEraseTime);
    }
}

static asynStatus
eraseSector(struct flashInfo *f, asynUser *pasynUser, unsigned int address)
{
    int i;
    int retry;

    asynPrint(pasynUser, ASYN_TRACE_FLOW, "Erase %#x\n", address);
    if (address >= f->capacity) {
        printf("Erase address beyond end of flash.\n");
        return asynError;
    }
    for (retry = 0 ;;) {
        if (tryToErase(f, pasynUser, address) == 0)
            break;
        if (++retry >= 3) {
            epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                                                    "Unrecoverable error.");
             return asynError;
        }
    }
    for (i = 0 ; i < 4 ; i++) {
        if ((address >= f->eraseBlockInfo[i].base)
         && (address < f->eraseBlockInfo[i].base +
                      f->eraseBlockInfo[i].size * f->eraseBlockInfo[i].count)) {
            f->eraseHighWater = address -
                                (address%f->eraseBlockInfo[i].size) +
                                f->eraseBlockInfo[i].size;
            break;
        }
    }
    asynPrint(pasynUser, ASYN_TRACE_FLOW, "Erased %#x to %#x\n", address, f->eraseHighWater);
    return asynSuccess;
}

static asynStatus
tryToProgram(struct flashInfo *f, asynUser *pasynUser, unsigned int address, epicsUInt16 value)
{
    int pass, passLimit = 1000;
    unsigned int d;
    int key;

    key = epicsInterruptLock();
    writeFlash(f, f->byteMode ? 0xAAA : 0x555<<1, 0xAA);
    writeFlash(f, f->byteMode ? 0x555 : 0xAAA<<1, 0x55);
    writeFlash(f, f->byteMode ? 0xAAA : 0x555<<1, 0xA0);
    writeFlash(f, address, value);
    epicsInterruptUnlock(key);
    for(pass = 0 ; ; pass++) {
        d = readFlash(f, address);
        if ((d & 0x80) == (value & 0x80))
            return asynSuccess;
        if (pass > passLimit) {
            writeFlash(f, 0x0, 0xF0);
            epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                    "Timed out waiting for chip (address %#x readback %#x).",
                                                                    address, d);
            return asynError;
        }
        if ((d & 0x20) != 0) {
            d = readFlash(f, address);
            if ((d & 0x80) == (value & 0x80))
                return asynSuccess;
            writeFlash(f, 0x0, 0xF0);
            epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                        "Chip timed out (pass %d, address %#x, readback %#x)",
                                                            pass, address, d);
            return asynError;
        }
    }
}

static asynStatus
cfiProgram(void *dpvt, asynUser *pasynUser, unsigned int address, void *buf, int count)
{
    struct flashInfo *f = dpvt;
    epicsUInt16 value, rvalue;
    int retry;

    asynPrint(pasynUser, ASYN_TRACE_FLOW, "Program count %d base %#x\n", count, address);
    if (count == 0) {
        f->eraseHighWater = 0;
        return asynSuccess;
    }
    while (count > 0) {
        if (address >= f->eraseHighWater)
            if (eraseSector(f, pasynUser, address) < 0)
                return asynError;
        if (address >= f->eraseHighWater) {
            epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                                  "Erase didn't clear the right location.");
            return asynError;
        }
        if (f->byteMode)
            value = *(epicsUInt8 *)buf;
        else
            value = *(epicsUInt16 *)buf;
        for (retry = 0 ;;) {
            if (tryToProgram(f, pasynUser, address, value) == 0)
                break;
            if (++retry >= 3) {
                epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                                                     "Unrecoverable error.");
                 return asynError;
            }
        }
        rvalue =  readFlash(f, address);
        if (rvalue != value) {
            epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                                             "Wrote %#x to %#x, but read %#x.",
                                                        value, address, rvalue);
             return asynError;
        }
        if (f->byteMode) {
            address += sizeof(epicsUInt8);
            buf = (epicsUInt8 *)buf + 1;
            count -= sizeof(epicsUInt8);
        }
        else {
            address += sizeof(epicsUInt16);
            buf = (epicsUInt16 *)buf + 1;
            count -= sizeof(epicsUInt16);
        }
    }
    return asynSuccess;
}

static struct drvFlashBurnerMethods methods = {
   cfiReport,
   cfiProgram
};

static int
drvCFIFlashBurnerConfigure(const char *name, int base, int isIndirect, int bits)
{
    epicsUInt16 dummy;
    struct flashInfo *f;

    if ((bits != 8) && (bits != 16)) {
        printf("Flash width must be 8 bits or 16 bits.\n");
        return -1;
    }
    if (devReadProbe(bits/8, (void *)base, &dummy) != 0) {
        printf("Can't access memory at location %#x\n", base);
        return -1;
    }
    f = callocMustSucceed(1, sizeof *f, "drvCFIFlashBurnerConfigure");
    f->base = (void *)base;
    f->indirect = isIndirect;
    f->byteMode = (bits == 8);
    if (getFlashInfo(f, 0) != asynSuccess)
        return -1;
    if (devFlashBurnerRegister(name, &methods, f) < 0)
        return -1;
    return 0;
}

/*
 * IOC shell command registration
 */
#include <iocsh.h>
static const iocshArg drvCFIFlashBurnerConfigureArg0 = { "name",iocshArgString};
static const iocshArg drvCFIFlashBurnerConfigureArg1 = { "address",iocshArgInt};
static const iocshArg drvCFIFlashBurnerConfigureArg2 = { "indirect",iocshArgInt};
static const iocshArg drvCFIFlashBurnerConfigureArg3 = { "width",iocshArgInt};
static const iocshArg *drvCFIFlashBurnerConfigureArgs[] = {
    &drvCFIFlashBurnerConfigureArg0, &drvCFIFlashBurnerConfigureArg1,
    &drvCFIFlashBurnerConfigureArg2, &drvCFIFlashBurnerConfigureArg3};
static const iocshFuncDef drvCFIFlashBurnerConfigureFuncDef =
                  {"drvCFIFlashBurnerConfigure",4,drvCFIFlashBurnerConfigureArgs};
static void drvCFIFlashBurnerConfigureCallFunc(const iocshArgBuf *args)
{
    drvCFIFlashBurnerConfigure(args[0].sval, args[1].ival,
                               args[2].ival, args[3].ival);
}
static void
drvCFIFlashBurnerRegisterCommands(void)
{
    iocshRegister(&drvCFIFlashBurnerConfigureFuncDef,drvCFIFlashBurnerConfigureCallFunc);
}
epicsExportRegistrar(drvCFIFlashBurnerRegisterCommands);
